download_geo <- function(gse_id, gpl_id = NA) {
  require(GEOquery)

  message("INFO: Fetching study", gse_id, "from GEO")
  gse_nnn <- paste0(substr(gse_id, 1, nchar(gse_id)-3), "nnn")
  wget_string <- paste0("wget -r -np -nH --cut-dirs=4 ftp://ftp.ncbi.nlm.nih.gov/geo/series/", gse_nnn, "/", gse_id, "/matrix/")
  system(wget_string, intern = TRUE) 
  
  n_platforms <- length(list.files("matrix"))
  message("INFO: Number of platforms: ", n_platforms)
  gpl_id_index <- 1

  if(n_platforms > 1){
    message("INFO: Multi-platform study detected")
    if(is.na(gpl_id)| gpl_id == " "){
      message("...WARNING: Platform GPL ID not provided. Getting first platform by default")
      gpl_id_index <- 1
    }
    else{
      gpl_id_index <- get_platform_index_matrix_dir("matrix/", gpl_id)
      if(gpl_id_index == 0)
        stop("ERROR: ", gpl_id, " does not belong to this study. Enter correct platform")
      message("...INFO: Fetching matrix for platform ", gpl_id)
    }
  }
  message("INFO: Parsing matrix")
  geo_object <- getGEO(filename = paste0("matrix/", list.files("matrix/")[[gpl_id_index]]), getGPL = F)
  return(geo_object)
}


get_platform_index_matrix_dir <- function(matrix_dir_path, gpl_id){
  require(GEOquery)
  file_choose_idx = 0 
  for(file_idx in 1:length(list.files(matrix_dir_path))){
    file <- list.files(matrix_dir_path)[file_idx]
    invisible(capture.output(file_platform <- getGEO(filename = paste0("matrix/", file), getGPL = F, GSEMatrix = F)@annotation))
    if(file_platform == gpl_id)
      file_choose_idx = file_idx
  }
  return(file_choose_idx)
}


download_gpl <- function(gpl_id) {
  require(GEOquery)
  if (!file.exists(paste("soft/", gpl_id, ".txt", sep = ""))){
    message("INFO: Downloading data for platform ", gpl_id)
    gseurl <- "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi"
    platform_url <- paste(gseurl, "?targ=self&acc=", gpl_id, "&form=text&view=full", sep = "")
    wget_string <- paste("mkdir soft/; wget -O soft/", gpl_id, ".txt ", "\"", platform_url, "\"", sep = "")
    system(wget_string, intern = T)
  }
  gpl_object <- getGEO(filename = paste("soft/", gpl_id, ".txt", sep = ""))
  return(gpl_object)
}

study_metadata_from_geo_object <- function(gse_gpl_object){
  study_meta_df <- data.frame(c("gse_id" = gse_gpl_object@experimentData@other$geo_accession,
                     "organism" = paste(unique(gse_gpl_object$organism_ch1), collapse = ","),
                     "title" = gse_gpl_object@experimentData@title,
                     "overall_design" = gse_gpl_object@experimentData@other$overall_design,
                     "summary" = gse_gpl_object@experimentData@other$summary,
                     "pubmedIds" = paste(strsplit(gse_gpl_object@experimentData@pubMedIds, "\n")[[1]], collapse = ",")))
  colnames(study_meta_df) <- c("property")
  return(study_meta_df)
}

study_metadata_from_geo_object_list <- function(gse_gpl_object, input_gpl){
  study_meta_list <- list("gse_id" = gse_gpl_object@experimentData@other$geo_accession,
                          "dataset_id"= paste(gse_gpl_object@experimentData@other$geo_accession, input_gpl, sep = "_"), 
                          "organism" = paste(unique(gse_gpl_object$organism_ch1), collapse = ","),
                          "title" = gse_gpl_object@experimentData@title,
                          "overall_design" = gse_gpl_object@experimentData@other$overall_design,
                          "summary" = gse_gpl_object@experimentData@other$summary,
                          "pubmedIds" = strsplit(gse_gpl_object@experimentData@pubMedIds, "\n")[[1]])
  return(study_meta_list)
}


add_from_annot <- function(gpl_object, package_name, annot_object_name){
  message(paste0("INFO: adding from ", package_name))
  eval(parse(text = paste0("require(",package_name, ")")))
  annot_obj <- eval(parse(text = annot_object_name))
  mapped_probes <- mappedkeys(annot_obj)
  # Convert to a list
  mapped_probes_list <- as.list(annot_obj)[mapped_probes]
  entrez_mapped <- unlist(mapped_probes_list[gpl_object@dataTable@table$ID])
  gpl_object@dataTable@table <- cbind(gpl_object@dataTable@table, entrez_mapped[gpl_object@dataTable@table$ID])
  colnames(gpl_object@dataTable@table) <- c(colnames(gpl_object@dataTable@table)[1:(ncol(gpl_object@dataTable@table)-1)], "ENTREZ_NEW")
  annot_column <- "ENTREZ_NEW"
  return(gpl_object)
}


rowname_from_gpl <- function(gse_matrix, gpl_object, annot_column,annot_type){
  gse_matrix_df <- as.data.frame(gse_matrix)
  if (class(rownames(gse_matrix))=="character"){
      match_positions <- match(rownames(gse_matrix), as.character(gpl_object@dataTable@table$ID))
  } else{
      match_positions <- match(rownames(gse_matrix), gpl_object@dataTable@table$ID)
  }


  if (grepl("HTA-2_0", gpl_object@header$title)){
    gpl_object <- add_from_annot(gpl_object, "hta20probeset.db", "hta20probesetENTREZID")
    annot_column <- "ENTREZ_NEW"
    annot_type <- "entrez"
  }


  if(grepl("Clariom_S_Human_HT", gpl_object@header$title)){
    gpl_object <- add_from_annot(gpl_object, "clariomshumanhttranscriptcluster.db", "clariomshumanhttranscriptclusterENTREZID")
    annot_column <- "ENTREZ_NEW"
    annot_type <- "entrez" 
  } 
  if(grepl("Clariom_S_Mouse", gpl_object@header$title)){
    gpl_object <- add_from_annot(gpl_object, "clariomsmousetranscriptcluster.db", "clariomsmousetranscriptclusterENTREZID")
    annot_column <- "ENTREZ_NEW"
    annot_type <- "entrez"
  }
  if(grepl("Clariom_S_Human", gpl_object@header$title)){
    gpl_object <- add_from_annot(gpl_object, "clariomdhumantranscriptcluster.db", "clariomdhumantranscriptclusterENTREZID")
    annot_column <- "ENTREZ_NEW"
    annot_type <- "entrez"
  } 
  
  
  if(grepl("Agilent-028005 SurePrint", gpl_object@header$title)){
    gpl_object <- add_from_annot(gpl_object, "MmAgilentDesign026655.db", "MmAgilentDesign026655ENTREZID")
    annot_column <- "ENTREZ_NEW"
    annot_type <- "entrez"
  }
    
  col=strsplit(annot_column, split=".1")[[1]][1]
  print(col)
  if (sum(!is.na(gpl_object@dataTable@table[,col])) == 0){
    stop("INFO: annotation column is empty in GPL file.")
  }
    
  gse_matrix_df$probe_id <- gpl_object@dataTable@table[, "ID"][match_positions]
    
  if (annot_column=="ENTREZ_NEW"){
    gse_matrix_df$pl_annot <- gpl_object@dataTable@table[, annot_column][match_positions]
  } else{
      if (col %in% colnames(gpl_object@dataTable@table)){
        gse_matrix_df$pl_annot <- gpl_object@dataTable@table[, col][match_positions]
      } else{
        print(annot_column)
        gse_matrix_df$pl_annot <- gpl_object@dataTable@table[, annot_column][match_positions]
      }
  }
    
  gse_matrix_df <- gse_matrix_df[!is.na(gse_matrix_df$pl_annot), ]
  gse_matrix_df <- gse_matrix_df[gse_matrix_df$pl_annot != "", ]
  results=list(df=gse_matrix_df,ann=annot_type)
  return(results)
}

remove_duplicate_symbols_mad <- function(gse_matrix_df){

  if (ncol(gse_matrix_df)>3){
    gse_matrix_df$mad_score <- apply(
    gse_matrix_df[, !grepl("symbol|pl_annot|probe_id", colnames(gse_matrix_df))],
    1, function(x) mad(x, na.rm = T))
    gse_matrix_df <- gse_matrix_df[order(gse_matrix_df$mad_score, decreasing = T), ]
  }

  gse_matrix_df <- gse_matrix_df[!duplicated(gse_matrix_df$symbol), ]
  rownames(gse_matrix_df) <- gse_matrix_df$symbol
  gse_matrix_df <- gse_matrix_df[, !grepl("symbol|pl_annot|mad_score", colnames(gse_matrix_df)), drop=FALSE]
  return(gse_matrix_df)
}

annotate_gse_matrix <- function(gse_matrix, platform_name, annot_column, annot_type) {
  require(AnnoProbe)
  gpl_obj <- download_gpl(platform_name)
  message(gpl_obj@dataTable@columns)

  if (nrow(gpl_obj@dataTable@columns)==0)
    stop('ERROR: GPL table is empty')

  source(paste(path, "/conversion.R", sep =""))

  message("INFO: column for annotation- ", annot_column)
  message("INFO: annotation type- ", annot_type)

  
  # identify entrez gene column

  results <- rowname_from_gpl(gse_matrix, gpl_obj, annot_column,annot_type)
  gse_matrix_df <- results[[1]]
  annot_type <- results[[2]]
  message("INFO: column for annotation- ", annot_column)
  message("INFO: annotation type- ", annot_type)

  if (annot_type == "entrez") {
    gse_matrix_df <- entrez_to_symbol(gse_matrix_df)
  }
  
  if (annot_type == "probe2gene") {
    gse_matrix_df <- probe_to_gene(gse_matrix_df, platform_name)
  }

  if (annot_type == "symbol") {
    gse_matrix_df <- entrez_to_symbol(gse_matrix_df, col_to_convert = "pl_annot", symbol_to_symbol = TRUE)
  }

  if (annot_type == "ensembl") {
    gse_matrix_df <- ensembl_to_symbol(gse_matrix_df)
  }

  if (annot_type == "gene_ass") {
    gse_matrix_df <- gene_assign_to_symbol(gse_matrix_df)
  }

  if (annot_type == "assoc_gene") {
    gse_matrix_df <- gene_assoc_to_symbol(gse_matrix_df)
  }

  if (annot_type == "genbank") {
    gse_matrix_df <- gb_acc_to_symbol(gse_matrix_df)
  }

  if (annot_type == "genbank_gblist") {
    gse_matrix_df <- gb_list_to_symbol(gse_matrix_df)
  }

  gse_matrix <- remove_duplicate_symbols_mad(gse_matrix_df)
  return(gse_matrix)
}

check_log2 <- function(gse_matrix) {
  gse_matrix <- gse_matrix[,!grepl("probe_id|entrez",colnames(gse_matrix)), drop=FALSE]
  gse_matrix[] <- sapply(gse_matrix, as.numeric)
  gse_matrix <- as.matrix(gse_matrix)
  if (abs(range(gse_matrix)[2]) - abs(range(gse_matrix)[1]) > 100) {
    return(FALSE)
  } else {
    return(TRUE)
  }
}

tranform_log2 <- function(gse_matrix) {
  gse_matrix_trans <- gse_matrix[,!grepl("probe_id|entrez",colnames(gse_matrix)), drop=FALSE]
  gse_matrix_trans <- as.matrix(gse_matrix_trans)
  gene_annot <- gse_matrix[,c("probe_id","entrez")]
  gene_annot$id <- 1:nrow(gene_annot)

  if (sum(gse_matrix_trans <= 0) > 0) {
    message("WARNING: The dataset has negative values which is weird")
    message("INFO: Make negative values and zeros as 1")
    gse_matrix_trans[gse_matrix_trans <= 0 ] <- 1
  }
  gse_matrix_trans <- log2(gse_matrix_trans)
  # merge and reorder
  gse_matrix_trans <- transform(merge(gse_matrix_trans, gene_annot, by='row.names'), row.names=Row.names, Row.names=NULL)
  gse_matrix_trans <- gse_matrix_trans[order(gse_matrix_trans$id),]
  gse_matrix_trans <- subset(gse_matrix_trans, select = -c(id))
  return(gse_matrix_trans)
 
}
